/*
 *
 * Copyright (C) 2009 Tomas Mazanec
 *
 * Author:   Tomas Mazanec <mazanec at utia.cas.cz>
 * Created:  Fri Aug 28 10:16:54 CEST 2009
 *
 * Application of CUDA in DSP algorithms 
 *	- Cross Ambiguity Function (CAF) implementation
 *
 */

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <unistd.h>
#include <time.h>

// includes, project
#include <cutil_inline.h>
#include <cufft.h>

#include <cucomplex.cu>

// includes, kernels
#include <cutesto.cu>
#include <tCAF_kernel.cu>

// forward

// Complex data type
typedef float2 Complex; 
// Complex math
static __device__ __host__ inline Complex ComplexConj(Complex);
static __device__ __host__ inline Complex ComplexAdd(Complex, Complex);
static __device__ __host__ inline Complex ComplexScale(Complex, float);
static __device__ __host__ inline Complex ComplexMul(Complex, Complex);
static __global__ void ComplexPointwiseMulAndScale(Complex*, const Complex*, int, float);

// definitions

#define MAXTHREADS 512
#define MAXSHMEMSIZE 16384

////////////////////////////////////////////////////////////////////////////////
// declaration, forward
void runTest(int argc, char** argv);


////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char** argv) 
{
    runTest(argc, argv);

    cutilExit(argc, argv);
}

////////////////////////////////////////////////////////////////////////////////
// Run test for CUDA
////////////////////////////////////////////////////////////////////////////////
void runTest(int argc, char** argv) 
{
    int dev = 0;

    // use command-line specified CUDA device, otherwise use device with highest Gflops/s
    if( cutCheckCmdLineFlag(argc, (const char**)argv, "device") )
	cutilDeviceInit(argc, argv);

    else{ cutilSafeCall(cudaSetDevice(dev)); }

    // kick the rand() if needed
//    srand( (unsigned int)time(NULL) );


    const char *fpaths1="../../../projects/tCAF/cafdata/s1cplx.dat";
    const char *fpaths2="../../../projects/tCAF/cafdata/s2cplx.dat";

    const unsigned int N = 131072;
    const unsigned int Nlog2N = 17;
    const unsigned int m = 600;
    const unsigned int L = 1; //loop with 120 data-arrays only

    Complex *h_sig1, *h_sig2;
    Complex *d_sig1, *d_sig2;
    Complex *h_sigt, *d_sigt;

    const unsigned int s1size = sizeof(Complex) * N;
    const unsigned int s2size = sizeof(Complex) * (N+m);

    const unsigned int stlen1x  = N;
    const unsigned int stlen    = N * m;
    const unsigned int stlendev = N * m/L; 
//    const unsigned int stsize1x = sizeof(Complex) * N;
    const unsigned int stsize   = sizeof(Complex) * N * m;
    const unsigned int stsizedev= sizeof(Complex) * N * m/L;
    
    const unsigned int NT = MAXTHREADS;
    const unsigned int NB = (unsigned int)(
				(unsigned int)N/ (unsigned int)NT);

    // 16384Bytes/8Bytes 
    const unsigned int shmem_len  = 3*512; // MAXSHMEMSIZE/sizeof(Complex);
    const unsigned int shmem_size = shmem_len*sizeof(Complex);

    // no. of kernel loops // 1 for now
    const unsigned int NL = 1;//shmem_len/NT;

    printf("CAF Exec: N:%d, tau 0->%d, with batch size: %d batches %d-times looped.\n",
	 N, m-1, m/L, L);
    
    printf("Kernel Exec: NT:%d, NB:%d, shmem_len:%d of Complex; NL:%d\n",
	 NT, NB, shmem_len, NL);

    
    // input data, 2 cplx vectors
    h_sig1 = (Complex *) malloc(s1size);
    h_sig2 = (Complex *) malloc(s2size);

    cutilCheckError( readFileComplexF(fpaths1, h_sig1, N, CUTFalse) );
    cutilCheckError( readFileComplexF(fpaths2, h_sig2, N, CUTFalse) );

    // complex conjugate sig2
    unsigned int timertest = 0;
    cutilCheckError( cutCreateTimer( &timertest));
    cutilCheckError( cutStartTimer( timertest));
    for (int i=0; i<N;i++) {
	h_sig2[i].y = (float) (-1.0 * h_sig2[i].y);
    }
    cutilCheckError( cutStopTimer( timertest));
    float timetest = cutGetTimerValue( timertest);
    cutilCheckError( cutDeleteTimer( timertest));

    float flopsytest = (N/(timetest/1000)) / 1e6;
    printf( "Host code of Conjugate in %f (ms) and %e ops => %f MFlops.\n",
	    timetest, ((float)N), flopsytest);
    

    // copy all to device                              
    cutilSafeCall(cudaMalloc((void**)&d_sig1, s1size));
    cutilSafeCall(cudaMemcpy(d_sig1, h_sig1, s1size, cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMalloc((void**)&d_sig2, s2size));
    cutilSafeCall(cudaMemcpy(d_sig2, h_sig2, s2size, cudaMemcpyHostToDevice));

    // prepare output
    h_sigt = (Complex *) malloc(stsize);
    cutilSafeCall(cudaMalloc((void**)&d_sigt, stsizedev));


    // prepare kernel
    dim3 grid(NB, 1, 1);
    dim3 block(NT, 1, 1);

    // and base ptrs for loop
    Complex *h_sigt_base = h_sigt;
    Complex *d_sig2_base = d_sig2;
    Complex *d_sigt_base = d_sigt;

    // prepare FFT
    cufftHandle plan;
    Complex *h_sigff;
    // and base ptr
    h_sigff = (Complex *) malloc(stsize);
    Complex *h_sigff_base = h_sigff;
    
    cufftSafeCall(cufftPlan1d(&plan, N, CUFFT_C2C, m/L) );

    printf("Check sizeof cufftComplex:%dB vs. Complex:%dB \n", sizeof(cufftComplex) , sizeof(Complex));
    printf("stsizedev:%dB , fftdatasize:%dB\n", stsizedev, sizeof(cufftComplex)*stlendev);


    // paranoid
    cutilCheckError( cutDeleteTimer( timertest));
    // global timer
    unsigned int timer = 0;
    cutilCheckError( cutCreateTimer( &timer));
    cutilCheckError( cutStartTimer( timer));

////////////////////////////////////////////////////////////////////////////////
// BIG LOOP
////////////////////////////////////////////////////////////////////////////////

  // prepare    
  unsigned int tau = 0;
    
  for ( unsigned int loopCounter=0; loopCounter < L; loopCounter++)
  {

////////////////////////////////////////////////////////////////////////////////
// kernel exec
////////////////////////////////////////////////////////////////////////////////

    for (unsigned int k=0; k < m/L; k++)
    {
	tau = k + loopCounter*m/L;
	d_sig2 = d_sig2_base + tau; // increment ptr to sig2 by 1
	d_sigt = d_sigt_base + k*stlen1x; // incr. ptr to sigt by 1 output

        // do the sig1 .* sig2 sliding by tau
	cafMulKernel <<< grid, block, shmem_size >>>(d_sig1, d_sig2, d_sigt, N, m, NL);
    }
    d_sigt = d_sigt_base;

    // check if kernel execution generated and error // when _DEBUG also calls cudaThreadSynchronize() 
    cutilCheckMsg("Kernel execution failed");
    // wait till kernel is done
    cutilSafeCall(cudaThreadSynchronize() );

#ifdef _DEBUG
    h_sigt = h_sigt_base + loopCounter * stlendev;
    // get results of m/L multiplcations // sig1 .* sig2(n+tau)
    cutilSafeCall( cudaMemcpy( h_sigt, d_sigt, stsizedev, cudaMemcpyDeviceToHost) );
    cutilSafeCall(cudaThreadSynchronize() );
    
#endif

////////////////////////////////////////////////////////////////////////////////
// FFT part
////////////////////////////////////////////////////////////////////////////////

    cufftSafeCall(cufftExecC2C(plan, (cufftComplex *)d_sigt, (cufftComplex *)d_sigt, CUFFT_FORWARD) );

    // wait till threads are done
    cutilSafeCall(cudaThreadSynchronize() );

    h_sigff = h_sigff_base + loopCounter * stlendev;
    cutilSafeCall( cudaMemcpy( (Complex *)h_sigff, (Complex *)d_sigt, stsizedev, cudaMemcpyDeviceToHost) );



////////////////////////////////////////////////////////////////////////////////
// END of all computations
////////////////////////////////////////////////////////////////////////////////

  }

////////////////////////////////////////////////////////////////////////////////
// END of BIG LOOP
////////////////////////////////////////////////////////////////////////////////


    cutilCheckError( cutStopTimer( timer));
    float timeker = 0;
    timeker = cutGetTimerValue( timer);
    cutilCheckError( cutDeleteTimer( timer));

    float num_ops = (m*N*Nlog2N) + (m*N);
    float flopsy = (num_ops/ (timeker/1000) ) / 1e6;
    printf( "--->>>\nOverall Processing time: %f (ms) and Ops: %e\n MegaFlopsy: %f\n\n",
	    timeker, num_ops, flopsy);


#ifdef _DEBUG
    h_sigt = h_sigt_base;
    if (m < 50)
    {
	cutilCheckError( writeFileComplexF("../../../projects/tCAF/cafdata/outputmul.dat", h_sigt, stlen ));
    }
    else {printf("Skipping out file outputmul.dat; m > 50\n");}
#endif

    h_sigff = h_sigff_base;
    if (m < 50)
    {
        cutilCheckError( writeFileComplexF("../../../projects/tCAF/cafdata/outputfft.dat", h_sigff, stlen ));
    }
    else {printf("Skipping out file outputfft.dat; m > 50\n");}
    

    // clean-up
    cufftSafeCall( cufftDestroy(plan) );

    h_sigt = h_sigt_base;
    h_sigff = h_sigff_base;
    d_sig2 = d_sig2_base;

    // clean-up
    free(h_sig1);
    free(h_sig2);
    free(h_sigt);
    free(h_sigff);
    cutilSafeCall(cudaFree(d_sig1));
    cutilSafeCall(cudaFree(d_sig2));
    cutilSafeCall(cudaFree(d_sigt));
    
    cudaThreadExit();
    
////////////////////////////////////////////////////////////////////////////////
// EOF 
////////////////////////////////////////////////////////////////////////////////

}

//

